@@ -155,16 +155,6 @@ set_props(const struct libevdev *dev, int fd, struct uinput_user_dev *uidev)
 	return rc;
 }
 
-static int
-open_uinput(void)
-{
-	int fd = open("/dev/uinput", O_RDWR|O_CLOEXEC);
-	if (fd < 0)
-		return -errno;
-
-	return fd;
-}
-
 LIBEVDEV_EXPORT int
 libevdev_uinput_get_fd(const struct libevdev_uinput *uinput_dev)
 {
@@ -189,7 +179,8 @@ fetch_device_node(const char *path)
 	/* ndev should only ever be 1 */
 
 	for (i = 0; i < ndev; i++) {
-		asprintf(&devnode, "/dev/input/%s", namelist[i]->d_name);
+		if (!devnode && asprintf(&devnode, "/dev/input/%s", namelist[i]->d_name) == -1)
+			devnode = NULL;
 		free(namelist[i]);
 	}
 
@@ -277,14 +268,15 @@ libevdev_uinput_create_from_device(const struct libevdev *dev, int fd, struct li
 		return -ENOMEM;
 
 	if (fd == LIBEVDEV_UINPUT_OPEN_MANAGED) {
-		fd = open_uinput();
+		fd = open("/dev/uinput", O_RDWR|O_CLOEXEC);
 		if (fd < 0)
-			return fd;
+			goto error;
 
 		new_device->fd_is_managed = 1;
 	} else if (fd < 0) {
 		log_bug("Invalid fd %d\n", fd);
-		return -EBADF;
+		errno = EBADF;
+		goto error;
 	}
 
 	memset(&uidev, 0, sizeof(uidev));
@@ -342,6 +334,9 @@ error:
 LIBEVDEV_EXPORT void
 libevdev_uinput_destroy(struct libevdev_uinput *uinput_dev)
 {
+	if (!uinput_dev)
+		return;
+
 	ioctl(uinput_dev->fd, UI_DEV_DESTROY, NULL);
 	if (uinput_dev->fd_is_managed)
 		close(uinput_dev->fd);
