@@ -36,23 +36,64 @@
 
 #define MAXEVENTS 64
 
-/* DEPRECATED */
-LIBEVDEV_EXPORT const enum libevdev_read_flag LIBEVDEV_READ_SYNC = LIBEVDEV_READ_FLAG_SYNC;
-LIBEVDEV_EXPORT const enum libevdev_read_flag LIBEVDEV_READ_NORMAL = LIBEVDEV_READ_FLAG_NORMAL;
-LIBEVDEV_EXPORT const enum libevdev_read_flag LIBEVDEV_FORCE_SYNC = LIBEVDEV_READ_FLAG_FORCE_SYNC;
-LIBEVDEV_EXPORT const enum libevdev_read_flag LIBEVDEV_READ_BLOCKING = LIBEVDEV_READ_FLAG_BLOCKING;
+enum event_filter_status {
+	EVENT_FILTER_NONE,	/**< Event untouched by filters */
+	EVENT_FILTER_MODIFIED,	/**< Event was modified */
+	EVENT_FILTER_DISCARD,	/**< Discard current event */
+};
 
 static int sync_mt_state(struct libevdev *dev, int create_events);
 
+static inline int*
+slot_value(const struct libevdev *dev, int slot, int axis)
+{
+	if (unlikely(slot > dev->num_slots)) {
+		log_bug("Slot %d exceeds number of slots (%d)\n", slot, dev->num_slots);
+		slot = 0;
+	}
+	if (unlikely(axis < ABS_MT_MIN || axis > ABS_MT_MAX)) {
+		log_bug("MT axis %d is outside the valid range [%d,%d]\n",
+			axis, ABS_MT_MIN, ABS_MT_MAX);
+		axis = ABS_MT_MIN;
+	}
+	return &dev->mt_slot_vals[slot * ABS_MT_CNT + axis - ABS_MT_MIN];
+}
+
 static int
 init_event_queue(struct libevdev *dev)
 {
-	/* FIXME: count the number of axes, keys, etc. to get a better idea at how many events per
-	   EV_SYN we could possibly get. Then multiply that by the actual buffer size we care about */
+	const int MIN_QUEUE_SIZE = 256;
+	int nevents = 1; /* terminating SYN_REPORT */
+	int nslots;
+	unsigned int type, code;
 
-	const int QUEUE_SIZE = 256;
+	/* count the number of axes, keys, etc. to get a better idea at how
+	   many events per EV_SYN we could possibly get. That's the max we
+	   may get during SYN_DROPPED too. Use double that, just so we have
+	   room for events while syncing an event.
+	 */
+	for (type = EV_KEY; type < EV_MAX; type++) {
+		int max = libevdev_event_type_get_max(type);
+		for (code = 0; max > 0 && code < (unsigned int) max; code++) {
+			if (libevdev_has_event_code(dev, type, code))
+				nevents++;
+		}
+	}
 
-	return queue_alloc(dev, QUEUE_SIZE);
+	nslots = libevdev_get_num_slots(dev);
+	if (nslots > 1) {
+		int num_mt_axes = 0;
+
+		for (code = ABS_MT_SLOT; code < ABS_MAX; code++) {
+			if (libevdev_has_event_code(dev, EV_ABS, code))
+				num_mt_axes++;
+		}
+
+		/* We already counted the first slot in the initial count */
+		nevents += num_mt_axes * (nslots - 1);
+	}
+
+	return queue_alloc(dev, max(MIN_QUEUE_SIZE, nevents * 2));
 }
 
 static void
@@ -102,7 +143,7 @@ log_msg(enum libevdev_log_priority priority,
 {
 	va_list args;
 
-	if (!log_data.handler)
+	if (!log_data.handler || priority > log_data.priority)
 		return;
 
 	va_start(args, format);
@@ -113,6 +154,13 @@ log_msg(enum libevdev_log_priority priority,
 static void
 libevdev_reset(struct libevdev *dev)
 {
+	free(dev->name);
+	free(dev->phys);
+	free(dev->uniq);
+	free(dev->mt_slot_vals);
+	free(dev->mt_sync.mt_state);
+	free(dev->mt_sync.tracking_id_changes);
+	free(dev->mt_sync.slot_update);
 	memset(dev, 0, sizeof(*dev));
 	dev->fd = -1;
 	dev->initialized = false;
@@ -128,7 +176,7 @@ libevdev_new(void)
 {
 	struct libevdev *dev;
 
-	dev = malloc(sizeof(*dev));
+	dev = calloc(1, sizeof(*dev));
 	if (!dev)
 		return NULL;
 
@@ -161,21 +209,11 @@ libevdev_free(struct libevdev *dev)
 	if (!dev)
 		return;
 
-	free(dev->name);
-	free(dev->phys);
-	free(dev->uniq);
 	queue_free(dev);
+	libevdev_reset(dev);
 	free(dev);
 }
 
-/* DEPRECATED */
-LIBEVDEV_EXPORT void
-libevdev_set_log_handler(struct libevdev *dev, libevdev_log_func_t logfunc)
-{
-	/* Can't be backwards compatible to this yet, so don't even try */
-	fprintf(stderr, "libevdev: ABI change. Log function will not be honored.\n");
-}
-
 LIBEVDEV_EXPORT void
 libevdev_set_log_function(libevdev_log_func_t logfunc, void *data)
 {
@@ -347,16 +385,47 @@ libevdev_set_fd(struct libevdev* dev, int fd)
 				goto out;
 
 			dev->abs_info[i] = abs_info;
-			if (i == ABS_MT_SLOT) {
-				dev->num_slots = abs_info.maximum + 1;
-				dev->current_slot = abs_info.value;
-			}
-
 		}
 	}
 
 	dev->fd = fd;
-	sync_mt_state(dev, 0);
+
+	/* devices with ABS_MT_SLOT - 1 aren't MT devices,
+	   see the documentation for multitouch-related
+	   functions for more details */
+	if (!libevdev_has_event_code(dev, EV_ABS, ABS_MT_SLOT - 1) &&
+	    libevdev_has_event_code(dev, EV_ABS, ABS_MT_SLOT)) {
+		const struct input_absinfo *abs_info;
+
+		abs_info = libevdev_get_abs_info(dev, ABS_MT_SLOT);
+
+		dev->num_slots = abs_info->maximum + 1;
+		dev->mt_slot_vals = calloc(dev->num_slots * ABS_MT_CNT, sizeof(int));
+		if (!dev->mt_slot_vals) {
+			rc = -ENOMEM;
+			goto out;
+		}
+		dev->current_slot = abs_info->value;
+
+		dev->mt_sync.mt_state_sz = sizeof(*dev->mt_sync.mt_state) +
+					   (dev->num_slots) * sizeof(int);
+		dev->mt_sync.mt_state = calloc(1, dev->mt_sync.mt_state_sz);
+
+		dev->mt_sync.tracking_id_changes_sz = NLONGS(dev->num_slots) * sizeof(long);
+		dev->mt_sync.tracking_id_changes = malloc(dev->mt_sync.tracking_id_changes_sz);
+
+		dev->mt_sync.slot_update_sz = NLONGS(dev->num_slots * ABS_MT_CNT) * sizeof(long);
+		dev->mt_sync.slot_update = malloc(dev->mt_sync.slot_update_sz);
+
+		if (!dev->mt_sync.tracking_id_changes ||
+		    !dev->mt_sync.slot_update ||
+		    !dev->mt_sync.mt_state) {
+			rc = -ENOMEM;
+			goto out;
+		}
+
+	    sync_mt_state(dev, 0);
+	}
 
 	rc = init_event_queue(dev);
 	if (rc < 0) {
@@ -509,27 +578,32 @@ out:
 static int
 sync_mt_state(struct libevdev *dev, int create_events)
 {
+	struct input_event *ev;
+	struct input_absinfo abs_info;
 	int rc;
-	int i;
+	int axis, slot;
 	int ioctl_success = 0;
-	struct mt_state {
-		int code;
-		int val[MAX_SLOTS];
-	} mt_state[ABS_MT_CNT];
+	int last_reported_slot = 0;
+	struct mt_sync_state *mt_state = dev->mt_sync.mt_state;
+	unsigned long *slot_update = dev->mt_sync.slot_update;
+	unsigned long *tracking_id_changes = dev->mt_sync.tracking_id_changes;
+	int need_tracking_id_changes = 0;
 
-	memset(&mt_state, 0, sizeof(mt_state));
+	memset(dev->mt_sync.slot_update, 0, dev->mt_sync.slot_update_sz);
+	memset(dev->mt_sync.tracking_id_changes, 0,
+	       dev->mt_sync.tracking_id_changes_sz);
 
-	for (i = ABS_MT_MIN; i <= ABS_MT_MAX; i++) {
-		int idx;
-		if (i == ABS_MT_SLOT)
+#define AXISBIT(_slot, _axis) (_slot * ABS_MT_CNT + _axis - ABS_MT_MIN)
+
+	for (axis = ABS_MT_MIN; axis <= ABS_MT_MAX; axis++) {
+		if (axis == ABS_MT_SLOT)
 			continue;
 
-		if (!libevdev_has_event_code(dev, EV_ABS, i))
+		if (!libevdev_has_event_code(dev, EV_ABS, axis))
 			continue;
 
-		idx = i - ABS_MT_MIN;
-		mt_state[idx].code = i;
-		rc = ioctl(dev->fd, EVIOCGMTSLOTS(sizeof(struct mt_state)), &mt_state[idx]);
+		mt_state->code = axis;
+		rc = ioctl(dev->fd, EVIOCGMTSLOTS(dev->mt_sync.mt_state_sz), mt_state);
 		if (rc < 0) {
 			/* if the first ioctl fails with -EINVAL, chances are the kernel
 			   doesn't support the ioctl. Simply continue */
@@ -537,70 +611,165 @@ sync_mt_state(struct libevdev *dev, int create_events)
 				rc = 0;
 			} else /* if the second, ... ioctl fails, really fail */
 				goto out;
-		} else if (ioctl_success == 0)
-			ioctl_success = 1;
-	}
+		} else {
+			if (ioctl_success == 0)
+				ioctl_success = 1;
 
-	for (i = 0; i < dev->num_slots; i++) {
-		int j;
-		struct input_event *ev;
+			for (slot = 0; slot < dev->num_slots; slot++) {
 
-		if (create_events) {
-			ev = queue_push(dev);
-			init_event(dev, ev, EV_ABS, ABS_MT_SLOT, i);
-		}
+				if (*slot_value(dev, slot, axis) == mt_state->val[slot])
+					continue;
 
-		for (j = ABS_MT_MIN; j <= ABS_MT_MAX; j++) {
-			int jdx = j - ABS_MT_MIN;
+				if (axis == ABS_MT_TRACKING_ID &&
+				    *slot_value(dev, slot, axis) != -1 &&
+				    mt_state->val[slot] != -1) {
+					set_bit(tracking_id_changes, slot);
+					need_tracking_id_changes = 1;
+				}
 
-			if (j == ABS_MT_SLOT)
-				continue;
+				*slot_value(dev, slot, axis) = mt_state->val[slot];
 
-			if (!libevdev_has_event_code(dev, EV_ABS, j))
-				continue;
-
-			if (dev->mt_slot_vals[i][jdx] == mt_state[jdx].val[i])
-				continue;
-
-			if (create_events) {
-				ev = queue_push(dev);
-				init_event(dev, ev, EV_ABS, j, mt_state[jdx].val[i]);
+				set_bit(slot_update, AXISBIT(slot, axis));
+				/* note that this slot has updates */
+				set_bit(slot_update, AXISBIT(slot, ABS_MT_SLOT));
 			}
-			dev->mt_slot_vals[i][jdx] = mt_state[jdx].val[i];
+
+
 		}
 	}
 
+	if (!create_events) {
+		rc = 0;
+		goto out;
+	}
+
+	if (need_tracking_id_changes) {
+		for (slot = 0; slot < dev->num_slots;  slot++) {
+			if (!bit_is_set(tracking_id_changes, slot))
+				continue;
+
+			ev = queue_push(dev);
+			init_event(dev, ev, EV_ABS, ABS_MT_SLOT, slot);
+			ev = queue_push(dev);
+			init_event(dev, ev, EV_ABS, ABS_MT_TRACKING_ID, -1);
+
+			last_reported_slot = slot;
+		}
+
+		ev = queue_push(dev);
+		init_event(dev, ev, EV_SYN, SYN_REPORT, 0);
+	}
+
+	for (slot = 0; slot < dev->num_slots;  slot++) {
+		if (!bit_is_set(slot_update, AXISBIT(slot, ABS_MT_SLOT)))
+			continue;
+
+		ev = queue_push(dev);
+		init_event(dev, ev, EV_ABS, ABS_MT_SLOT, slot);
+		last_reported_slot = slot;
+
+		for (axis = ABS_MT_MIN; axis <= ABS_MT_MAX; axis++) {
+			if (axis == ABS_MT_SLOT ||
+			    !libevdev_has_event_code(dev, EV_ABS, axis))
+				continue;
+
+			if (bit_is_set(slot_update, AXISBIT(slot, axis))) {
+				ev = queue_push(dev);
+				init_event(dev, ev, EV_ABS, axis, *slot_value(dev, slot, axis));
+			}
+		}
+	}
+
+	/* add one last slot event to make sure the client is on the same
+	   slot as the kernel */
+
+	rc = ioctl(dev->fd, EVIOCGABS(ABS_MT_SLOT), &abs_info);
+	if (rc < 0)
+		goto out;
+
+	dev->current_slot = abs_info.value;
+
+	if (dev->current_slot != last_reported_slot) {
+		ev = queue_push(dev);
+		init_event(dev, ev, EV_ABS, ABS_MT_SLOT, dev->current_slot);
+	}
+
+#undef AXISBIT
+
 	rc = 0;
 out:
 	return rc ? -errno : 0;
 }
 
+static int
+read_more_events(struct libevdev *dev)
+{
+	int free_elem;
+	int len;
+	struct input_event *next;
+
+	free_elem = queue_num_free_elements(dev);
+	if (free_elem <= 0)
+		return 0;
+
+	next = queue_next_element(dev);
+	len = read(dev->fd, next, free_elem * sizeof(struct input_event));
+	if (len < 0) {
+		return -errno;
+	} else if (len > 0 && len % sizeof(struct input_event) != 0)
+		return -EINVAL;
+	else if (len > 0) {
+		int nev = len/sizeof(struct input_event);
+		queue_set_num_elements(dev, queue_num_elements(dev) + nev);
+	}
+
+	return 0;
+}
+
+static inline void
+drain_events(struct libevdev *dev)
+{
+	int rc;
+	size_t nelem;
+	int iterations = 0;
+	const int max_iterations = 8; /* EVDEV_BUF_PACKETS in
+					 kernel/drivers/input/evedev.c */
+
+	queue_shift_multiple(dev, queue_num_elements(dev), NULL);
+
+	do {
+		rc = read_more_events(dev);
+		if (rc == -EAGAIN)
+			return;
+
+		if (rc < 0) {
+			log_error("Failed to drain events before sync.\n");
+			return;
+		}
+
+		nelem = queue_num_elements(dev);
+		queue_shift_multiple(dev, nelem, NULL);
+	} while (iterations++ < max_iterations && nelem >= queue_size(dev));
+
+	/* Our buffer should be roughly the same or bigger than the kernel
+	   buffer in most cases, so we usually don't expect to recurse. If
+	   we do, make sure we stop after max_iterations and proceed with
+	   what we have.  This could happen if events queue up faster than
+	   we can drain them.
+	 */
+	if (iterations >= max_iterations)
+		log_info("Unable to drain events, buffer size mismatch.\n");
+}
+
 static int
 sync_state(struct libevdev *dev)
 {
-	int i;
 	int rc = 0;
 	struct input_event *ev;
 
-	/* FIXME: if we have events in the queue after the SYN_DROPPED (which was
-	   queue[0]) we need to shift this backwards. Except that chances are that the
-	   queue may be either full or too full to prepend all the events needed for
-	   SYNC_IN_PROGRESS.
-
-	   so we search for the last sync event in the queue and drop everything before
-	   including that event and rely on the kernel to tell us the right value for that
-	   bitfield during the sync process.
-	 */
-
-	for (i = queue_num_elements(dev) - 1; i >= 0; i--) {
-		struct input_event e = {{0,0}, 0, 0, 0};
-		queue_peek(dev, i, &e);
-		if (e.type == EV_SYN)
-			break;
-	}
-
-	if (i > 0)
-		queue_shift_multiple(dev, i + 1, NULL);
+	 /* see section "Discarding events before synchronizing" in
+	  * libevdev/libevdev.h */
+	drain_events(dev);
 
 	if (libevdev_has_event_type(dev, EV_KEY))
 		rc = sync_key_state(dev);
@@ -610,7 +779,8 @@ sync_state(struct libevdev *dev)
 		rc = sync_sw_state(dev);
 	if (rc == 0 && libevdev_has_event_type(dev, EV_ABS))
 		rc = sync_abs_state(dev);
-	if (rc == 0 && libevdev_has_event_code(dev, EV_ABS, ABS_MT_SLOT))
+	if (rc == 0 && dev->num_slots > -1 &&
+	    libevdev_has_event_code(dev, EV_ABS, ABS_MT_SLOT))
 		rc = sync_mt_state(dev, 1);
 
 	dev->queue_nsync = queue_num_elements(dev);
@@ -641,20 +811,20 @@ update_key_state(struct libevdev *dev, const struct input_event *e)
 static int
 update_mt_state(struct libevdev *dev, const struct input_event *e)
 {
-	if (e->code == ABS_MT_SLOT) {
+	if (e->code == ABS_MT_SLOT && dev->num_slots > -1) {
 		int i;
 		dev->current_slot = e->value;
 		/* sync abs_info with the current slot values */
 		for (i = ABS_MT_SLOT + 1; i <= ABS_MT_MAX; i++) {
 			if (libevdev_has_event_code(dev, EV_ABS, i))
-				dev->abs_info[i].value = dev->mt_slot_vals[dev->current_slot][i - ABS_MT_MIN];
+				dev->abs_info[i].value = *slot_value(dev, dev->current_slot, i);
 		}
 
 		return 0;
 	} else if (dev->current_slot == -1)
 		return 1;
 
-	dev->mt_slot_vals[dev->current_slot][e->code - ABS_MT_MIN] = e->value;
+	*slot_value(dev, dev->current_slot, e->code) = e->value;
 
 	return 0;
 }
@@ -732,35 +902,47 @@ update_state(struct libevdev *dev, const struct input_event *e)
 	return rc;
 }
 
-static int
-read_more_events(struct libevdev *dev)
+/**
+ * Sanitize/modify events where needed.
+ */
+static inline enum event_filter_status
+sanitize_event(const struct libevdev *dev,
+	       struct input_event *ev,
+	       enum SyncState sync_state)
 {
-	int free_elem;
-	int len;
-	struct input_event *next;
+	if (unlikely(dev->num_slots > -1 &&
+		     libevdev_event_is_code(ev, EV_ABS, ABS_MT_SLOT) &&
+		     (ev->value < 0 || ev->value >= dev->num_slots))) {
+		log_bug("Device \"%s\" received an invalid slot index %d."
+				"Capping to announced max slot number %d.\n",
+				dev->name, ev->value, dev->num_slots - 1);
+		ev->value = dev->num_slots - 1;
+		return EVENT_FILTER_MODIFIED;
 
-	free_elem = queue_num_free_elements(dev);
-	if (free_elem <= 0)
-		return 0;
-
-	next = queue_next_element(dev);
-	len = read(dev->fd, next, free_elem * sizeof(struct input_event));
-	if (len < 0) {
-		return -errno;
-	} else if (len > 0 && len % sizeof(struct input_event) != 0)
-		return -EINVAL;
-	else if (len > 0) {
-		int nev = len/sizeof(struct input_event);
-		queue_set_num_elements(dev, queue_num_elements(dev) + nev);
+	/* Drop any invalid tracking IDs, they are only supposed to go from
+	   N to -1 or from -1 to N. Never from -1 to -1, or N to M. Very
+	   unlikely to ever happen from a real device.
+	   */
+	} else if (unlikely(sync_state == SYNC_NONE &&
+			    dev->num_slots > -1 &&
+			    libevdev_event_is_code(ev, EV_ABS, ABS_MT_TRACKING_ID) &&
+			    ((ev->value == -1 &&
+			     *slot_value(dev, dev->current_slot, ABS_MT_TRACKING_ID) == -1) ||
+			     (ev->value != -1 &&
+			     *slot_value(dev, dev->current_slot, ABS_MT_TRACKING_ID) != -1)))) {
+		log_bug("Device \"%s\" received a double tracking ID %d in slot %d.\n",
+			dev->name, ev->value, dev->current_slot);
+		return EVENT_FILTER_DISCARD;
 	}
 
-	return 0;
+	return EVENT_FILTER_NONE;
 }
 
 LIBEVDEV_EXPORT int
 libevdev_next_event(struct libevdev *dev, unsigned int flags, struct input_event *ev)
 {
 	int rc = LIBEVDEV_READ_STATUS_SUCCESS;
+	enum event_filter_status filter_status;
 
 	if (!dev->initialized) {
 		log_bug("device not initialized. call libevdev_set_fd() first\n");
@@ -793,14 +975,13 @@ libevdev_next_event(struct libevdev *dev, unsigned int flags, struct input_event
 		   of the device too */
 		while (queue_shift(dev, &e) == 0) {
 			dev->queue_nsync--;
-			update_state(dev, &e);
+			if (sanitize_event(dev, &e, dev->sync_state) != EVENT_FILTER_DISCARD)
+				update_state(dev, &e);
 		}
 
 		dev->sync_state = SYNC_NONE;
 	}
 
-	/* FIXME: if the first event after SYNC_IN_PROGRESS is a SYN_DROPPED, log this */
-
 	/* Always read in some more events. Best case this smoothes over a potential SYN_DROPPED,
 	   worst case we don't read fast enough and end up with SYN_DROPPED anyway.
 
@@ -825,12 +1006,15 @@ libevdev_next_event(struct libevdev *dev, unsigned int flags, struct input_event
 		if (queue_shift(dev, ev) != 0)
 			return -EAGAIN;
 
-		update_state(dev, ev);
+		filter_status = sanitize_event(dev, ev, dev->sync_state);
+		if (filter_status != EVENT_FILTER_DISCARD)
+			update_state(dev, ev);
 
 	/* if we disabled a code, get the next event instead */
-	} while(!libevdev_has_event_code(dev, ev->type, ev->code));
+	} while(filter_status == EVENT_FILTER_DISCARD ||
+		!libevdev_has_event_code(dev, ev->type, ev->code));
 
-	rc = 0;
+	rc = LIBEVDEV_READ_STATUS_SUCCESS;
 	if (ev->type == EV_SYN && ev->code == SYN_DROPPED) {
 		dev->sync_state = SYNC_NEEDED;
 		rc = LIBEVDEV_READ_STATUS_SYNC;
@@ -839,8 +1023,15 @@ libevdev_next_event(struct libevdev *dev, unsigned int flags, struct input_event
 	if (flags & LIBEVDEV_READ_FLAG_SYNC && dev->queue_nsync > 0) {
 		dev->queue_nsync--;
 		rc = LIBEVDEV_READ_STATUS_SYNC;
-		if (dev->queue_nsync == 0)
+		if (dev->queue_nsync == 0) {
+			struct input_event next;
 			dev->sync_state = SYNC_NONE;
+
+			if (queue_peek(dev, 0, &next) == 0 &&
+			    next.type == EV_SYN && next.code == SYN_DROPPED)
+				log_info("SYN_DROPPED received after finished "
+					 "sync - you're not keeping up\n");
+		}
 	}
 
 out:
@@ -893,9 +1084,9 @@ LIBEVDEV_EXPORT void libevdev_set_##field(struct libevdev *dev, const char *fiel
 	dev->field = strdup(field); \
 }
 
-STRING_SETTER(name);
-STRING_SETTER(phys);
-STRING_SETTER(uniq);
+STRING_SETTER(name)
+STRING_SETTER(phys)
+STRING_SETTER(uniq)
 
 
 #define PRODUCT_GETTER(name) \
@@ -904,10 +1095,10 @@ LIBEVDEV_EXPORT int libevdev_get_id_##name(const struct libevdev *dev) \
 	return dev->ids.name; \
 }
 
-PRODUCT_GETTER(product);
-PRODUCT_GETTER(vendor);
-PRODUCT_GETTER(bustype);
-PRODUCT_GETTER(version);
+PRODUCT_GETTER(product)
+PRODUCT_GETTER(vendor)
+PRODUCT_GETTER(bustype)
+PRODUCT_GETTER(version)
 
 #define PRODUCT_SETTER(field) \
 LIBEVDEV_EXPORT void libevdev_set_id_##field(struct libevdev *dev, int field) \
@@ -915,10 +1106,10 @@ LIBEVDEV_EXPORT void libevdev_set_id_##field(struct libevdev *dev, int field) \
 	dev->ids.field = field;\
 }
 
-PRODUCT_SETTER(product);
-PRODUCT_SETTER(vendor);
-PRODUCT_SETTER(bustype);
-PRODUCT_SETTER(version);
+PRODUCT_SETTER(product)
+PRODUCT_SETTER(vendor)
+PRODUCT_SETTER(bustype)
+PRODUCT_SETTER(version)
 
 LIBEVDEV_EXPORT int
 libevdev_get_driver_version(const struct libevdev *dev)
@@ -951,7 +1142,7 @@ libevdev_has_event_type(const struct libevdev *dev, unsigned int type)
 LIBEVDEV_EXPORT int
 libevdev_has_event_code(const struct libevdev *dev, unsigned int type, unsigned int code)
 {
-	const unsigned long *mask;
+	const unsigned long *mask = NULL;
 	int max;
 
 	if (!libevdev_has_event_type(dev, type))
@@ -971,7 +1162,7 @@ libevdev_has_event_code(const struct libevdev *dev, unsigned int type, unsigned
 LIBEVDEV_EXPORT int
 libevdev_get_event_value(const struct libevdev *dev, unsigned int type, unsigned int code)
 {
-	int value;
+	int value = 0;
 
 	if (!libevdev_has_event_type(dev, type) || !libevdev_has_event_code(dev, type, code))
 		return 0;
@@ -981,6 +1172,19 @@ libevdev_get_event_value(const struct libevdev *dev, unsigned int type, unsigned
 		case EV_KEY: value = bit_is_set(dev->key_values, code); break;
 		case EV_LED: value = bit_is_set(dev->led_values, code); break;
 		case EV_SW: value = bit_is_set(dev->sw_values, code); break;
+		case EV_REP:
+			    switch(code) {
+				    case REP_DELAY:
+					    libevdev_get_repeat(dev, &value, NULL);
+					    break;
+				    case REP_PERIOD:
+					    libevdev_get_repeat(dev, NULL, &value);
+					    break;
+				    default:
+					    value = 0;
+					    break;
+			    }
+			    break;
 		default:
 			value = 0;
 			break;
@@ -1002,6 +1206,9 @@ libevdev_set_event_value(struct libevdev *dev, unsigned int type, unsigned int c
 	e.code = code;
 	e.value = value;
 
+	if (sanitize_event(dev, &e, SYNC_NONE) != EVENT_FILTER_NONE)
+		return -1;
+
 	switch(type) {
 		case EV_ABS: rc = update_abs_state(dev, &e); break;
 		case EV_KEY: rc = update_key_state(dev, &e); break;
@@ -1032,13 +1239,13 @@ libevdev_get_slot_value(const struct libevdev *dev, unsigned int slot, unsigned
 	if (!libevdev_has_event_type(dev, EV_ABS) || !libevdev_has_event_code(dev, EV_ABS, code))
 		return 0;
 
-	if (dev->num_slots < 0 || slot >= (unsigned int)dev->num_slots || slot >= MAX_SLOTS)
+	if (dev->num_slots < 0 || slot >= (unsigned int)dev->num_slots)
 		return 0;
 
 	if (code > ABS_MT_MAX || code < ABS_MT_MIN)
 		return 0;
 
-	return dev->mt_slot_vals[slot][code - ABS_MT_MIN];
+	return *slot_value(dev, slot, code);
 }
 
 LIBEVDEV_EXPORT int
@@ -1047,7 +1254,7 @@ libevdev_set_slot_value(struct libevdev *dev, unsigned int slot, unsigned int co
 	if (!libevdev_has_event_type(dev, EV_ABS) || !libevdev_has_event_code(dev, EV_ABS, code))
 		return -1;
 
-	if (dev->num_slots == -1 || slot >= (unsigned int)dev->num_slots || slot >= MAX_SLOTS)
+	if (dev->num_slots == -1 || slot >= (unsigned int)dev->num_slots)
 		return -1;
 
 	if (code > ABS_MT_MAX || code < ABS_MT_MIN)
@@ -1059,8 +1266,7 @@ libevdev_set_slot_value(struct libevdev *dev, unsigned int slot, unsigned int co
 		dev->current_slot = value;
 	}
 
-	dev->mt_slot_vals[slot][code - ABS_MT_MIN] = value;
-
+	*slot_value(dev, slot, code) = value;
 
 	return 0;
 }
@@ -1071,7 +1277,7 @@ libevdev_fetch_slot_value(const struct libevdev *dev, unsigned int slot, unsigne
 	if (libevdev_has_event_type(dev, EV_ABS) &&
 	    libevdev_has_event_code(dev, EV_ABS, code) &&
 	    dev->num_slots >= 0 &&
-	    slot < (unsigned int)dev->num_slots && slot < MAX_SLOTS) {
+	    slot < (unsigned int)dev->num_slots) {
 		*value = libevdev_get_slot_value(dev, slot, code);
 		return 1;
 	} else
@@ -1107,11 +1313,11 @@ LIBEVDEV_EXPORT int libevdev_get_abs_##name(const struct libevdev *dev, unsigned
 	return absinfo ? absinfo->name : 0; \
 }
 
-ABS_GETTER(maximum);
-ABS_GETTER(minimum);
-ABS_GETTER(fuzz);
-ABS_GETTER(flat);
-ABS_GETTER(resolution);
+ABS_GETTER(maximum)
+ABS_GETTER(minimum)
+ABS_GETTER(fuzz)
+ABS_GETTER(flat)
+ABS_GETTER(resolution)
 
 #define ABS_SETTER(field) \
 LIBEVDEV_EXPORT void libevdev_set_abs_##field(struct libevdev *dev, unsigned int code, int val) \
@@ -1239,12 +1445,6 @@ libevdev_disable_event_code(struct libevdev *dev, unsigned int type, unsigned in
 	return 0;
 }
 
-LIBEVDEV_EXPORT int
-libevdev_kernel_set_abs_value(struct libevdev *dev, unsigned int code, const struct input_absinfo *abs)
-{
-	return libevdev_kernel_set_abs_info(dev, code, abs);
-}
-
 LIBEVDEV_EXPORT int
 libevdev_kernel_set_abs_info(struct libevdev *dev, unsigned int code, const struct input_absinfo *abs)
 {
@@ -1298,22 +1498,12 @@ libevdev_grab(struct libevdev *dev, enum libevdev_grab_mode grab)
 	return rc < 0 ? -errno : 0;
 }
 
-/* DEPRECATED */
-LIBEVDEV_EXPORT int
-libevdev_is_event_type(const struct input_event *ev, unsigned int type)
-ALIAS(libevdev_event_is_type);
-
 LIBEVDEV_EXPORT int
 libevdev_event_is_type(const struct input_event *ev, unsigned int type)
 {
 	return type < EV_CNT && ev->type == type;
 }
 
-/* DEPRECATED */
-LIBEVDEV_EXPORT int
-libevdev_is_event_code(const struct input_event *ev, unsigned int type, unsigned int code)
-ALIAS(libevdev_event_is_code);
-
 LIBEVDEV_EXPORT int
 libevdev_event_is_code(const struct input_event *ev, unsigned int type, unsigned int code)
 {
@@ -1326,11 +1516,6 @@ libevdev_event_is_code(const struct input_event *ev, unsigned int type, unsigned
 	return (max > -1 && code <= (unsigned int)max && ev->code == code);
 }
 
-/* DEPRECATED */
-LIBEVDEV_EXPORT const char*
-libevdev_get_event_type_name(unsigned int type)
-ALIAS(libevdev_event_type_get_name);
-
 LIBEVDEV_EXPORT const char*
 libevdev_event_type_get_name(unsigned int type)
 {
@@ -1340,11 +1525,6 @@ libevdev_event_type_get_name(unsigned int type)
 	return ev_map[type];
 }
 
-/* DEPRECATED */
-LIBEVDEV_EXPORT const char*
-libevdev_get_event_code_name(unsigned int type, unsigned int code)
-ALIAS(libevdev_event_code_get_name);
-
 LIBEVDEV_EXPORT const char*
 libevdev_event_code_get_name(unsigned int type, unsigned int code)
 {
@@ -1356,16 +1536,6 @@ libevdev_event_code_get_name(unsigned int type, unsigned int code)
 	return event_type_map[type][code];
 }
 
-/* DEPRECATED */
-LIBEVDEV_EXPORT const char*
-libevdev_get_input_prop_name(unsigned int prop)
-ALIAS(libevdev_property_get_name);
-
-/* DEPRECATED */
-LIBEVDEV_EXPORT const char*
-libevdev_get_property_name(unsigned int prop)
-ALIAS(libevdev_property_get_name);
-
 LIBEVDEV_EXPORT const char*
 libevdev_property_get_name(unsigned int prop)
 {
@@ -1375,11 +1545,6 @@ libevdev_property_get_name(unsigned int prop)
 	return input_prop_map[prop];
 }
 
-/* DEPRECATED */
-LIBEVDEV_EXPORT int
-libevdev_get_event_type_max(unsigned int type)
-ALIAS(libevdev_event_type_get_max);
-
 LIBEVDEV_EXPORT int
 libevdev_event_type_get_max(unsigned int type)
 {
@@ -1390,7 +1555,7 @@ libevdev_event_type_get_max(unsigned int type)
 }
 
 LIBEVDEV_EXPORT int
-libevdev_get_repeat(struct libevdev *dev, int *delay, int *period)
+libevdev_get_repeat(const struct libevdev *dev, int *delay, int *period)
 {
 	if (!libevdev_has_event_type(dev, EV_REP))
 		return -1;
