@@ -30,6 +30,8 @@ extern "C" {
 #include <linux/input.h>
 #include <stdarg.h>
 
+#define LIBEVDEV_ATTRIBUTE_PRINTF(_format, _args) __attribute__ ((format (printf, _format, _args)))
+
 /**
  * @mainpage
  *
@@ -544,7 +546,8 @@ extern "C" {
  * libevdev's internal test suite uses the
  * [Check unit testing framework](http://check.sourceforge.net/). Tests are
  * divided into test suites and test cases. Most tests create a uinput device,
- * so you'll need to run as root.
+ * so you'll need to run as root, and your kernel must have
+ * CONFIG_INPUT_UINPUT enabled.
  *
  * To run a specific suite only:
  *
@@ -789,7 +792,8 @@ typedef void (*libevdev_log_func_t)(enum libevdev_log_priority priority,
 				    void *data,
 				    const char *file, int line,
 				    const char *func,
-				    const char *format, va_list args);
+				    const char *format, va_list args)
+	LIBEVDEV_ATTRIBUTE_PRINTF(6, 0);
 
 /**
  * @ingroup init
